package jeliot.lang;

import java.lang.reflect.*;
import java.util.*;
import jeliot.parser.error.*;
import jeliot.FeatureNotImplementedException;

/**
  * @author Pekka Uronen
  *
  * created         27.9.1999
  */
public class ForeignType extends ReferenceType {

    private static Hashtable instanceIndex = new Hashtable();

    private Class myClass;

    private ForeignType() { }

    private ForeignType(Class c) {
        this.myClass = c;
    }

    /** Returns the instance of the type of given class.
     */
    public static ForeignType getType(Class c) {
        ForeignType ft = (ForeignType)instanceIndex.get(c);
        if (ft == null) {
            ft = new ForeignType(c);
            instanceIndex.put(c, ft);
        }
        return ft;
    }

    /** Checks if a value of this type may be assigned to a variable of
      * given type t.
      */
    public boolean mayBeAssignedTo(Type t) {
        // foreign-domestic assignment is not implemented yet.
        if (t instanceof DomesticType) {
            throw new FeatureNotImplementedException(
                "Mixing your own classes with compiled classes.");
        }
        if (! (t instanceof ForeignType) ) {
            return false;
        }
        // t must be a foreign type.
        ForeignType ft = (ForeignType)t;
        return ft.myClass.isAssignableFrom(myClass);
    }


    /** Checks if a value of this type can be cast to a value of type
      * t in an explicit cast.
      */
    public boolean mayBeCastTo(Type t) {
        // foreign-domestic assignment is not implemented yet.
        if (t instanceof DomesticType) {
            throw new FeatureNotImplementedException(
                "Mixing your own classes with compiled classes.");
        }
        if (! (t instanceof ForeignType) ) {
            return false;
        }
        // t must be a foreign type.
        ForeignType ft = (ForeignType)t;
        return ft.myClass.isAssignableFrom(myClass) ||
                myClass.isAssignableFrom(ft.myClass);
    }

    public boolean isInteface() {
        return myClass.isInterface();
    }

    public FieldPointer getField(String name) {
        try {
            Field field = myClass.getField(name);
            return new ForeignFieldPointer(field);
        }
        catch (NoSuchFieldException e) {
            throw new NonexistentFieldException(this, name);
        }    
    }
    
    /** Returns a constructor pointer to a constructor with arguments
      * of given types.
      */
    public ConstructorPointer getConstructor(Type[] argt) {
        
        Constructor constr[] = myClass.getConstructors();
        int n = constr.length;
        
        // Find applicable constructors
        Class[][] params = new Class[n][];
        for (int i = 0; i < n; ++i) {
            params[i] = constr[i].getParameterTypes();
        }
        Object[] accapp = (Object[])findApplicable(
                null, argt, null, params, constr);
        
        // Count the applicable constructors
        int count = 0;
        while (accapp[count] != null) {
            ++count;
        }
        if (count == 0) {
            throw new NonexistentConstructorException(argt);
        }
        
        // Choose the most specific constructor.
        if (count == 1) {
            return new ForeignConstructorPointer(
                    (Constructor)accapp[0]);
        }
        // most specific constructor this far.
        int most = 0;
        for (int i = 1; i < count; ++i) {
            if (isMoreSpecific((Constructor)accapp[i],
                    (Constructor)accapp[most])) {
                if (isMoreSpecific((Constructor)accapp[most], 
                        (Constructor)accapp[i])) {
                            
                    throw new AmbiguousMethodException(
                            (Constructor)accapp[most],
                            (Constructor)accapp[i]);
                }
                most = i;
            }
        }
        return new ForeignConstructorPointer(
                (Constructor)accapp[most]);
    }
    
    /** 15.11.2 */
    public MethodPointer getMethod(String methodName, Type[] argt) {

        // 15.11.2.1 -- Find methods that are applicable
        //              and accessible

        // methods of this class
        Method[] meths = myClass.getMethods();
        int n = meths.length;

        // Find applicable methods
        String[] names = new String[n];
        Class[][] params = new Class[n][];
        for (int i = 0; i < n; ++i) {
            names[i] = meths[i].getName();
            params[i] = meths[i].getParameterTypes();
        }
        Object[] accapp = findApplicable(
                methodName, argt, names, params, meths);
        
        // Count the applicable methods
        int count = 0;
        while (accapp[count] != null) {
            ++count;
        }
        if (count == 0) {
            throw new NonexistentMethodException(methodName, argt);
        }

        // 15.11.2.2 Choose the most specific method.
        if (count == 1) {
            return new ForeignMethodPointer((Method)accapp[0]);
        }
        // most specific method this far.
        int most = 0;
        for (int i = 1; i < count; ++i) {
            if (isMoreSpecific(
                    (Method)accapp[i], (Method)accapp[most])) {
                if (isMoreSpecific(
                        (Method)accapp[most], (Method)accapp[i])) {
                            
                    throw new AmbiguousMethodException(
                            (Method)accapp[most], (Method)accapp[i]);
                }
                most = i;
            }
        }
        return new ForeignMethodPointer((Method)accapp[most]);
    }
    
    private Object[] findApplicable(String methodName, Type[] argt,
            String[] names, Class[][] params, Object[] meths) {
              
        int n = meths.length;        
        // leave one extra space to a sentinel
        Object[] matching = new Object[n+1];        
        int count = 0;
        
        Outer: 
        for (int i = 0; i < n; ++i) {
            // check if names match
            if (methodName != null && !names[i].equals(methodName)) {
                continue Outer;
            }
            Class[] at = params[i];
            int m = at.length;

            // check if numbers of parameters match
            if (m != argt.length) {
                continue Outer;
            }
            if (m == 0) {
                // no arguments
                matching[count++] = meths[i];
                break Outer;
            }
            // count matching arguments
            Inner:
            for (int j = 0; j < m; ++j) {
                Type t = Type.getTypeByClass(at[j]);
                if (!argt[j].mayBeAssignedTo(t)) {
                    continue Outer;
                }
            }
            // all the arguments match
            matching[count++] = meths[i];
        }
        return matching;
    }
    
    /** Tests if method m1 is more specific than method m2.
      */
    private boolean isMoreSpecific(Method m1, Method m2) {
        Class c1 = m1.getDeclaringClass();
        Class c2 = m2.getDeclaringClass();
        if (!c2.isAssignableFrom(c1)) {
            return false;
        }
        Class[] p1 = m1.getParameterTypes();
        Class[] p2 = m2.getParameterTypes();
        return moreSpecificArguments(p1, p2);
    }

    /** Tests if Constructor m1 is more specific than constructor m2.
      */
    private boolean isMoreSpecific(Constructor m1, Constructor m2) {
        Class[] p1 = m1.getParameterTypes();
        Class[] p2 = m2.getParameterTypes();
        return moreSpecificArguments(p1, p2);
    }
    
    private boolean moreSpecificArguments(Class[] p1, Class[] p2) {
        int n = p1.length;
        for (int i = 0; i < n; ++i) {
            Type type1 = Type.getTypeByClass(p1[i]);
            Type type2 = Type.getTypeByClass(p2[i]);
            if (!type1.mayBeAssignedTo(type2)) {
                return false;
            }
        }
        return true;
    }
    
    public String toString() {
        return myClass.getName();
    }
    
    public String getSimpleName() {
        String name = myClass.getName();
        int index = name.lastIndexOf('.');
        return index == -1 ?
                name :
                name.substring(index+1);
    }
            
}

    