package jeliot.parser;

import java.util.*;
import jeliot.theatre.*;
import jeliot.lang.*;

/**
  * PFor represents a for statement.
  *
  * @author Pekka Uronen
  *
  * created         6.8.1999
  * last modified  15.8.1999
  */
public class PFor extends PStatement {

    private PExpression check;
    private PStatement statement;
    private PVariableDeclaration initDecl;
    private Vector initExpr;
    private Vector updateExpr;

    public PFor(
            PVariableDeclaration init, 
            PExpression check,
            Vector update,
            PStatement statement,
            int left, 
            int right) {

        super(left, right);
        this.check = check;
        this.statement = statement;
        this.initDecl = init;
        this.updateExpr = update;

		init.setParent(this);
		check.setParent(this);
		setParents(update);
		statement.setParent(this);
    }

    public PFor(
            Vector init, 
            PExpression check, 
            Vector update, 
            PStatement statement,
            int left,
            int right) {

        super(left, right);
        this.check = check;
        this.statement = statement;
        this.initExpr = init;
        this.updateExpr = update;

        setParents(init);
		check.setParent(this);
		setParents(update);
		statement.setParent(this);
    }

    public PExpression getCheck() {
        return check;
    }

    public PVariableDeclaration getDeclaration() {
        return initDecl;
    }

    public Enumeration getInitializers() {
        return initExpr == null ?
                null :
                initExpr.elements() ;
    }

    public Enumeration getUpdaters() {
        return updateExpr == null ?
                null :
                updateExpr.elements() ;
    }

    public void acceptVisitor(ProgramTreeVisitor visitor) {
        visitor.enterFor(this);
        if (initExpr != null) {
            Enumeration enum = initExpr.elements();
            while (enum.hasMoreElements()) {
                PExpression expr = (PExpression)enum.nextElement();
                expr.acceptVisitor(visitor);
            }
        }
        if (initDecl != null) {
            initDecl.acceptVisitor(visitor);
        }
        check.acceptVisitor(visitor);
        if (updateExpr != null) {
            Enumeration enum = updateExpr.elements();
            while (enum.hasMoreElements()) {
                PExpression expr = (PExpression)enum.nextElement();
                expr.acceptVisitor(visitor);
            }
        }
        statement.acceptVisitor(visitor);
        visitor.leaveFor(this);
    }

    public String statementName() {
        return "for";
    }
    
    public void execute(Director director) throws Return {
        // inform director we are at a new statement
        director.atStatement(this);

        // Evaluate initializers
        if (initExpr != null) {
            Enumeration enum = initExpr.elements();
            while (enum.hasMoreElements()) {
                PExpression expr = (PExpression)enum.nextElement();
                Scratch scratch = director.openScratch();
                expr.evaluate(director);
                director.closeScratch(scratch);
            }
        }
        else if (initDecl != null) {
            initDecl.execute(director);
            director.atStatement(this);
        }
                
        // evaluate the expression
        Scratch scratch = director.openScratch();
        Value v = check.evaluate(director);
        boolean contLoop = ((Boolean)v.getValue()).booleanValue();
        
        // inform director about the result
        if ( contLoop ) {
            director.enterLoop(this, v);
        }
        else {
            director.skipLoop(this, v);
        }
        director.closeScratch(scratch);
        
        while ( contLoop ) {
            statement.execute(director);
            
            // back to this statement
            director.atStatement(this);
            
            // evaluate the update expressions
            if (updateExpr != null) {
                Enumeration enum = updateExpr.elements();
                while (enum.hasMoreElements()) {
                    PExpression expr = (PExpression)enum.nextElement();
                    scratch = director.openScratch();
                    expr.evaluate(director);
                    director.closeScratch(scratch);
                }
            }
            
            // evaluate the expression
            scratch = director.openScratch();
            v = check.evaluate(director);
            contLoop = ((Boolean)v.getValue()).booleanValue();
            director.atStatement(this);
            if ( contLoop ) {
                director.continueLoop(this, v);
            }
            else {
                director.exitLoop(this, v);
            }
            director.closeScratch(scratch);
        }
    }
}
