package jeliot.parser;

import java.util.*;
import jeliot.lang.*;
import jeliot.theatre.*;

/**
  * PMethod class encapsulates a method.
  *
  * @author Pekka Uronen
  *
  * created         5.8.1999
  */
public class PMethod extends PElement implements PClassMember {

    private static Vector emptyVector = new Vector();

    /** Modifiers that apply to this method. (public, static, ...) */
    PModifiers modifiers;

    /** Return type of this method. Null means that it is a void method. */
    PType type;

    /** Name (identifier) of this method. */
    String name;

    /** Formal parameters of this method. Null means that there are none. */
    Vector parameters;

    /** Exceptions and errors thrown by this method. Null means that
      * there are none. */    
    Vector throwables;

    /** Dimension to be added to the dimension of this method's return type.
      * This makes it possible to declare method type in C-style. */
    int dimension;

    /** The body of this method. Null means that it is an abstract method. */ 
    PBlock body;
    
    /** The class this method is member of. */
    private PClass declaringClass;
    
    private Type myType;
    
    private DomesticMethodPointer pointer = 
            new DomesticMethodPointer(this);

    private int variableCount;
    
    public PMethod(String identifier, Vector parameters, int nameStart) {
        this.name = identifier;
        this.parameters = parameters;
        setLeftPos(nameStart);
        setParents(parameters);
    }
    
    public void setDeclaringClass(PClass declaringClass) {
        this.declaringClass = declaringClass;
    }
    
    public PClass getDeclaringClass() {
        return declaringClass;
    }
    
    public boolean isStatic() {
        return modifiers.hasModifier(PModifiers.STATIC);
    }

    /** Sets the header information of this method. This is called by the
      * parser. */
    public void setHeaderInformation(PModifiers modifiers, PType type,
            Vector throwables, int right) {

        this.modifiers = modifiers;
        this.type = type;
        this.throwables = throwables;
        setRightPos(right);
    }
    
    public String getIdentifier() {
        return name;
    }
    
    public void setBody(PBlock body) {
        this.body = body;
        if (body != null) {
            body.setParent(this);
        }
    }

    public void setMaximumCount(int count) {
        this.variableCount = count;
    }

    public int getMaximumCount() {
        return variableCount;
    }

    public int getLeftPos() {
        return modifiers.getLeftPos();
    }
    
    public PModifiers getModifiers() {
        return modifiers;
    }
    
    public Type getType() {
        return myType;
    }
    
    public Type resolveType(NameSpace space, PClass context) {
        if (parameters != null) {
            Enumeration enum = parameters.elements();
            while (enum.hasMoreElements()) {
                PVariableDeclaration dec = 
                    (PVariableDeclaration)enum.nextElement();
                dec.resolveType(space, context);
            }
        }
        
        PClass myClass = (PClass)getParent();
        return this.myType = type.resolveType(space, myClass);
    }
    
    
    /** Increases the dimension of this declarator by one. This is used
      * when constructing C-style array return type declarations. */
    public void incDimension() {
        ++dimension;
    }

    public String toString() {
        return "Method " + type + " " + name;
    }

    public void acceptVisitor(ProgramTreeVisitor visitor) throws SemanticException {
        
        // tell the visitor to handle this method
        visitor.visitMethod(this);

        // handle parameters
        if (parameters != null) {
            Enumeration enum = parameters.elements();
            while (enum.hasMoreElements()) {
                 PVariableDeclaration de = 
                        (PVariableDeclaration)enum.nextElement();
                 de.acceptVisitor(visitor);
            }
        }
        // recurse to method body
        body.acceptVisitor(visitor);

        visitor.leaveMethod(this);
    }
    
    public Enumeration getFormalParameters() {
        return new Enumeration() {
                    Enumeration enum = parameters == null ?
                            null : parameters.elements();
                    public Object nextElement() {
                        PVariableDeclaration d = 
                            (PVariableDeclaration)enum.nextElement();
                        return d.getDeclarators().nextElement();
                    }
                    public boolean hasMoreElements() {
                        return enum != null && enum.hasMoreElements();
                    }
                };
    }

    public int formalParameterCount() {
        return parameters == null   ?
                0                   :
                parameters.size()   ;
    }   

    public void execute(Director director) {
        execute(director, null);
    }
    
    public Return execute(Director director, Value[] args) {
        director.setUpMethod(this, args);
        try {
            body.execute(director);
            //director.bringDownMethod(this, null);
            return null;
        }
        catch (Return ret) {
            //director.bringDownMethod(this, ret);
            return ret;
        }
    }
    
    public String getName() {
        return name;
    }

    public MethodPointer getPointer() {
        return pointer;
    }
    
    /** Determines if this method's signature matches the given name and
      * arguments.
      */
    public boolean matches(String name, Type[] argt) {
        // check the name
        if (!this.name.equals(name)) {
            return false;
        }
        
        // check the argument list.
        if (parameters == null) {
            return argt.length == 0;
        }
        int m = parameters.size();
        if (argt.length != m) {
            return false;
        }
        // match arguments
        for (int i = 0; i < m; ++i) {
            PVariableDeclaration de = 
                    (PVariableDeclaration)parameters.elementAt(i);
            PVariableDeclarator d = (PVariableDeclarator)de.
                    getDeclarators().nextElement();
            Type t = d.getType();
            if (!argt[i].mayBeAssignedTo(t)) {
                return false;
            }
        }
        return true;
    }
    
}
