package jeliot.parser;

import jeliot.lang.*;
import jeliot.theatre.*;

/**
  * PUnaryExpr represents a unary expression (+, -, !, etc.).
  *
  * @author Pekka Uronen
  *
  * created         6.8.1999
  * last modified   15.8.1999
  */
public class PUnaryExpr extends PExpression {

    public static final int PLUS	        = 0;
    public static final int MINUS	        = 1;
    public static final int PLUSPLUS	    = 2;
    public static final int MINUSMINUS	    = 3;
    public static final int COMP	        = 4;
    public static final int NOT	            = 5;
    public static final int POSTPLUSPLUS	= 6;
    public static final int POSTMINUSMINUS	= 7;
    
    private int operator;
    private PExpression expression;

    public PUnaryExpr(int left, PExpression expression, int operator) {
        this.expression = expression;
        this.operator = operator;
        expression.setParent(this);
        setLeftPos(left);
        setRightPos(expression.getRightPos()); 
    }
    
    
    public PUnaryExpr(PExpression expression, int operator, 
            int right) {
        
        this.expression = expression;
        this.operator = operator;
        expression.setParent(this);
        setLeftPos(expression.getLeftPos());
        setRightPos(right); 
    }

    public int getOperator() {
        return operator;
    }

    public void acceptVisitor(ProgramTreeVisitor visitor) {
        visitor.enterUnaryExpression(this);
        expression.acceptVisitor(visitor);
        visitor.leaveUnaryExpression(this);
    }
    
    public PExpression getArgument() {
        return expression;
    }

    public Value evaluate(Director director) {
        super.evaluate(director);
        director.atExpression(this);
        switch (operator) {
            case (PLUS): {
                Value val = expression.evaluate(director);
                return evaluatePlus(director, val);
            }
            case (MINUS): {
                Value val = expression.evaluate(director);
                return evaluateMinus(director, val);
            }
            case (PLUSPLUS): {
                return evaluatePlusPlus(director, false);
            }
            case (MINUSMINUS): {
                return evaluateMinusMinus(director, false);
            }
            case (POSTPLUSPLUS): {
                return evaluatePlusPlus(director, true);
            }
            case (POSTMINUSMINUS): {
                return evaluateMinusMinus(director, true);
            }
            case (NOT): {
                Value val = expression.evaluate(director);
                boolean orig = ((Boolean)val.getValue()).booleanValue();
                Value newValue = Value.newBoolean(!orig); 
                director.animateUnaryExpression(NOT, val, newValue);
                return newValue;
            }
            default:
                throw new RuntimeException("Strange operator: "+operator);
        }
    }

    public Value evaluateMinus(Director director, Value value) {
        Type t = value.getType();
        Object val = value.getValue();
        Object result = null;
        
        if (val instanceof Character) {
            result = new Integer( -((Character)val).charValue() );
        }
        else if (t == PrimitiveType.longTYPE) {
            result = new Long( -((Long)val).longValue() );
        }
        else if (t.isIntegral()) {
            result = new Integer( -((Number)val).intValue() );
        }
        else if (t == PrimitiveType.floatTYPE) {
            result = new Float( -((Float)val).floatValue() );
        }
        else if (t == PrimitiveType.doubleTYPE) {
            result = new Double( -((Double)val).doubleValue() );
        }
        Value resVal = new Value(getType(), result);
        director.animateUnaryExpression(MINUS, value, resVal);
        return resVal;
    }

     public Value evaluatePlus(Director director, Value value) {
        Type t = getType();
        Object val = value.getValue();
        Object result = null;

        if (val instanceof Character) {
            result = new Integer( ((Character)val).charValue() );
        }
        else if (t == PrimitiveType.longTYPE) {
            result = new Long( ((Long)val).longValue() );
        }
        else if (t.isIntegral()) {
            result = new Integer( ((Number)val).intValue() );
        }
        else if (t == PrimitiveType.floatTYPE) {
            result = new Float( ((Float)val).floatValue() );
        }
        else if (t == PrimitiveType.doubleTYPE) {
            result = new Double( ((Double)val).doubleValue() );
        }
        Value resVal = new Value(getType(), result);
        director.animateUnaryExpression(PLUS, value, resVal);
        return resVal;
    }

    public Value evaluatePlusPlus(Director director, boolean post) {
        Variable var = ( (PVariableAccessExpr)expression ).
            evaluateVariable(director);
        Value value = var.getValue();
        Type t = getType();
        Object val = value.getValue();
        Object result = null;
        
        if (val instanceof Character) {
            result = new Integer( ((Character)val).charValue() +1);
        }
        else if (t == PrimitiveType.longTYPE) {
            result = new Long( ((Long)val).longValue() +1l);
        }
        else if (t.isIntegral()) {
            result = new Integer( ((Number)val).intValue() +1);
        }
        else if (t == PrimitiveType.floatTYPE) {
            result = new Float( ((Float)val).floatValue() +1.0f);
        }
        else if (t == PrimitiveType.doubleTYPE) {
            result = new Double( ((Double)val).doubleValue() +1.0);
        }
        Value resVal = new Value(getType(), result);
        var.assign(resVal);
        if (post) {
            PElement parent = getParent();
            director.animatePostIncDec(PLUSPLUS, var,
                    (parent instanceof PExpression ||
                     parent instanceof PVariableDeclarator) ?
                   value : null);
            return value;
        }
        else {
            director.animatePreIncDec(PLUSPLUS, var, resVal);
            return resVal;
        }
    }
        
    public Value evaluateMinusMinus(Director director, boolean post) {
        Variable var = ( (PVariableAccessExpr)expression ).
            evaluateVariable(director);
        Value value = var.getValue();
        Type t = getType();
        Object val = value.getValue();
        Object result = null;
        
        if (val instanceof Character) {
            result = new Integer( ((Character)val).charValue() -1);
        }
        else if (t == PrimitiveType.longTYPE) {
            result = new Long( ((Long)val).longValue() -1l);
        }
        else if (t.isIntegral()) {
            result = new Integer( ((Number)val).intValue() -1);
        }
        else if (t == PrimitiveType.floatTYPE) {
            result = new Float( ((Float)val).floatValue() -1.0f);
        }
        else if (t == PrimitiveType.doubleTYPE) {
            result = new Double( ((Double)val).doubleValue() -1.0);
        }
        Value resVal = new Value(getType(), result);
        var.assign(resVal);
        
        director.atExpression(this);
        if (post) {
            PElement parent = getParent();
            director.animatePostIncDec(MINUSMINUS, var,
                     (parent instanceof PExpression ||
                      parent instanceof PVariableDeclarator) ?
                     value : null);
            return value;
        }
        else {
            director.animatePreIncDec(MINUSMINUS, var, resVal);
            return resVal;
        }
    }

    public boolean usesScratch() {
        return !(operator == MINUSMINUS || operator == PLUSPLUS);
    }

}
