package jeliot.parser;

import jeliot.lang.*;
import jeliot.theatre.*;
import jeliot.parser.error.*;

/**
  * PVariableAccessExpr is produced by production
  *     postfix_expression --> name
  * or production
  *     left_hand_side --> name
  * It represents an access to a variable.
  *
  * @author Pekka Uronen
  *
  * created         6.8.1999
  */
public class PVariableAccessExpr extends PExpression {

    /** Name of the variable this expression accesses */
    private PName name;

    private PExpression access;
    private FieldPointer field;
 
    public PVariableAccessExpr(PName name) {
       super(name.getLeftPos(), name.getRightPos());
       this.name = name;     
    }

    public void acceptVisitor(ProgramTreeVisitor visitor) {
        visitor.visitVariableAccess(this);
    }
    
    public PName getName() {
        return name;
    }
    
    public PName getVariableName() {
        return name;
    }
    
    public String getIdentifier() {
        return name.getHead();
    }
    
    public Value evaluate(Director director) {
        super.evaluate(director);
        Variable variable = field.access(director);
        return director.initiateVariableAccess(variable);
    }

    public Variable evaluateVariable(Director director) {
        return field.access(director);
    }

    public boolean isPrimary() {
        return true;
    }

    public boolean usesScratch() {
        return false;
    }
    
    public Type resolveType(
            NameSpace space,
            LocalVariableBundle locals,
            PClass context,
            boolean staticContext) {
        
        if (name.isQualified()) {
            // this field access is of form le.ft.head
            PName left = name.getLeftHand();
            String head = name.getHead();
            left.classify(space, context);
            if (left.getClassification() == PName.TYPE) {
                // left hand side is a class name
                ReferenceType rt = space.findClass(left.toString());
                
                // set the referenced field  
                this.field = rt.getField(head);
                if (staticContext && !field.isStatic()) {
                    throw new StaticContextException(this);
                }
                // set the type of this expression as the type of
                // the referenced field.
                Type type = field.getType();
                setType(type);
                return type;
            }
            // left hand side is not a class name, so treat it as
            // a field access.
            this.access = new PVariableAccessExpr(left);
            access.setParent(this);
            
            // get the type of the access
            Type pt = ((PVariableAccessExpr)access).
                    resolveType(space, locals, context, staticContext);
            if (! (pt instanceof ReferenceType)) {
                 throw new IllegalFieldAccessException(pt, this);    
            }
            ReferenceType rt = (ReferenceType)pt;
                
            // get the referenced field
            try {
                this.field = rt.getField(head);
            }
            catch (NonexistentFieldException e) {
                e.setAccess(this);
                throw (e);
            }      
            
            // set the type of this expression as the type of
            // the referenced field.                    
            Type type = field.getType();
            setType(type);
            return type;                              
        }
        // simple name
            
        // get the name of the referenced field
        String id = name.getHead();
            
        //  Check if the access is to a local variable
        PVariableDeclarator vacc = (PVariableDeclarator)locals.get(id);
        if (vacc != null) {   
            // refering to a local variable
            
            // get the referenced variable
            this.field = vacc.getPointer();
            
            // set the type of this expression as the type of
            // the referenced variable.
            Type type = field.getType();
            setType(type);
            return type;
        }
        // not a local variable, so search the fields of 
        // enclosing class
        vacc = context.getField(id);
        if (vacc != null) {
            field = vacc.getPointer();
            Type type = field.getType();
            setType(type);
            return type;
        }
        NonexistentFieldException exc =  
                new NonexistentFieldException(context.getType(), id);
        exc.setAccess(this);
        throw (exc);
    }
    
}
